(async function() {
    'use strict';

    const parentExtensionId = document.currentScript.getAttribute('parentExtensionId');
    const parentExtensionUrlBase = 'chrome-extension://' + encodeURIComponent(parentExtensionId);
    // const util = await import('ic-util.mjs');
    const util = await import(parentExtensionUrlBase + '/ic-util.mjs');

    function InstagramVideoControls() {

        let showLog = false;

        /** @var {HTMLVideoElement} */
        let player;

        let config = {};

        const LOCALSTORAGE_VOLUME = "controls_for_instagram_volume";

        // true after .init() was called.
        let initialized = false;

        let mutationObserverEnabled = false;

        const knownVideoElements = new Set();

        function arr(a) {
            return Array.from(a);
        }

        function modifyVideos() {
            // log('modifyVideos()');

            // On 3rd party non-Instagram websites, they use a .EmbedVideo container to hold the video.
            // On Instagrams site, they don't.
            // So, we look for both styles.
            util.queryAll("video").forEach(video => {

                // Try to find a component root element, which is basically a div that's a few parents above the video element,
                // and serves as a container for all the instagram GUI stuff, like buttons, images, overlays, the video element etc...

                // .EmbedVideo is found when were in an iframe used by a 3rd party to embed instagram.
                let componentRoot = video.closest(".EmbedVideo");
                if (!componentRoot) {
                    // Try to find the style of root used on instagram.com when they overlay the video with a big play button.
                    componentRoot = util.recordEx('gevcre', getEstimatedVideoComponentRootElement)(video);
                }

                if (!componentRoot) {
                    // If we reached here, either they changed their dom structure, or its a video without the insta-junk markup. This seems to happen
                    // on very long videos - they just present a native html5 player, although they still disable downloads and other stuff.
                    // We'll just go up 4 parents like the other cases and hope for the best lol.
                    // I think we aren't likely to find any overlay buttons in this case, at least I havent encountered any yet.
                    // this should probably be the 5th parent instead of 4th, to be consistent /w our other code, but 4th works, and is safer.
                    componentRoot = nthParent(video, 4);

                    // Also, we need to mark the video as not having a component root.
                    // video.dataset.icNoClickToPlayHandlerNeeded = "1";
                }

                if (componentRoot) {
                    util.recordEx('mve', modifyVideoElement)(componentRoot);
                    util.recordEx('rowipc', removeOverlayWithInstagramPlayControl)(componentRoot);
                }

            });
        }

        function hide(el) {
            // Being visibility:hidden also prevents mouse events.
            // Also, a hidden element retains its size as determined by the bounding box.
            // Both of these factors matter, and other code depends on them being that way.
            el && el.style && (el.style.visibility = 'hidden');
        }

        /**
         * This element is their control which pauses the vid when you hold the mouse down.
         * It interferes with the <video> receiving mousemove events necessary to show the native controls,
         * and also interferes w/ normal click to toggle play/pause.
         *
         *
         * @param videoElement
         */
        function incapacitateStoryVideoPausingOverlays(videoElement) {
            // There's more than 1 element overlaying the video.
            // Also, they change this markup often.
            arr(videoElement.parentElement.children).filter(el => el !== videoElement).forEach(hide);
        }

        function modifyVideoElement(embedRootElem) {

            util.queryAll("video", embedRootElem).forEach(videoPlayer => {

                if (videoPlayer.dataset.nativeControlsISetJooFree) {
                    // Skip, we've already processed this element.
                    return;
                }

                // This is used as a css selector to redefine the style instatard uses to hide the native media controls.
                // We use the dataset instead of a class because the react framework that they use will overwrite some other properties, erasing our class name.
                videoPlayer.dataset.nativeControlsISetJooFree = "1";
                knownVideoElements.add(videoPlayer);

                // Enable native controls. Some instagram videos actually already have them enabled.
                videoPlayer.controls = true;

                // They tend to not show the download video option, so we reenable it by clearing the list of restricted controls.
                videoPlayer.setAttribute('controlsList', '');

                if (isInstagramStoriesPage()) {
                    incapacitateStoryVideoPausingOverlays(videoPlayer);
                }

                // Setup a click handler that only fires once.
                const clickHandler = evt => {
                    // We need to stop the event, otherwise clicking on the video element will open a new tab.
                    evt.stopPropagation();

                    // We only want to interfere on the first click, to start the video. It seems that after this first click,
                    // instagram installs its own click handlers that will play/pause the video, so we don't want to interfere and
                    // end up playing their pause, or pausing their play. But in some cases, they install a click handler before the first click, too.
                    // So, we can't be sure if instagram has a click to play handler already setup or not.
                    // If they do, we need to avoid calling play, or we may introduce a stutter or event loop.
                    // If we wait a small moment after the click, and then check again to see if the video hasn't started playing,
                    // we will assume they dont have a handler, because the vid should be playing if they did. Even if the network is slow,
                    // the video.paused attribute should immediately be false when .play() gets called.
                    if (!videoStartedPlaying(videoPlayer)) {
                        setTimeout(util.recordEx('play', () => {
                            if (!videoStartedPlaying(videoPlayer)) {
                                videoPlayer.play();
                            }
                        }), 50);
                    }

                    log(`removing click handler from inside click handler. im in ur click handler.`);
                    // We also remove the event handler in case the video is rewound to 0.00 time, in which case an event loop can occur.
                    videoPlayer.removeEventListener('click', util.recordEx('ch', clickHandler));
                };

                if (!videoPlayer.dataset.icNoClickToPlayHandlerNeeded) {
                    videoPlayer.addEventListener('click', clickHandler);
                }

                videoPlayer.addEventListener('volumechange', evt => {
                    if (config.rememberVolumeLevel && getSavedVolumeLevel() !== videoPlayer.volume) {
                        saveVolumeLevel(videoPlayer.volume);
                        setVolumeOfPreviouslySeenVideoElements(videoPlayer.volume);
                        log('The volume changed.', evt, videoPlayer.volume);
                    }
                });

                // We set the volume immediately because this is the first time we've seen this element.
                const savedVolumeLevel = getSavedVolumeLevel();
                if (config.rememberVolumeLevel) {
                    setVolumeIfChanged(videoPlayer, savedVolumeLevel);
                }

                setPlaybackRateIfChanged(videoPlayer, config.playbackRate);

                // We probably only need to add css to the page once, but im worried react might remove it in some obscure
                // case, so adding it over and over seems safer.
                redefineWebkitMediaControlHidingCssRule();
            });

            try {
                // This func has a high chance of failing w/ NPE if instagram changes their dom layout.
                // It's not critical for this function to succeed, so we just keep going if it throws.
                modifyVideoHeightIfSendMessageBoxIsBlockingVideoControls(embedRootElem);
            } catch (ex) {
                log(ex);
            }
        }

        function videoStartedPlaying(videoPlayer) {
            // A video will start paused, and as soon as a user clicks play (or video.play() is called) .paused will be false.
            // Even if the video hasn't started playing yet due to a slow download.

            // We assume it played if the time is not at the start. This should be true in the case of instagram, as I dont know of any case where they
            // start playback in the middle.
            return videoPlayer.currentTime > 0.0 || !videoPlayer.paused;
        }

        function setPlaybackRateIfChanged(player, newPlaybackRate) {
            // Only update if the current playbackRate is different from the new playbackRate (avoids triggering a new ratechange event).
            if (player && !isNaN(newPlaybackRate) && valuesAreDifferentEnough(player.playbackRate,  newPlaybackRate)) {
                try {
                    player.playbackRate = newPlaybackRate;
                } catch (ex) {
                    log('failed to set playback rate. it was probably out of range.', {player, playbackRate: newPlaybackRate, ex});
                }
            }
        }

        function setVolumeIfChanged(player, newVolume) {
            // Only update if the current volume is different from the new volume (avoids triggering a new volumechange event).
            if (player && !isNaN(newVolume) && valuesAreDifferentEnough(player.volume,  newVolume)) {
                try {
                    log('before set vol');
                    player.volume = newVolume;
                    // If you notice sometimes videos autoplay but start muted, it's likely this Chrome feature: https://developers.google.com/web/updates/2017/09/autoplay-policy-changes
                } catch (ex) {
                    log('failed to set volume. it was probably out of range.', {player, volume: newVolume, ex});
                }
            }
        }

        function setVolumeOfPreviouslySeenVideoElements(volume) {
            getKnownVideoElements().forEach(video => {
                setVolumeIfChanged(video, volume);
            });
        }

        function setPlaybackRateOfPreviouslySeenVideoElements(playbackRate) {
            getKnownVideoElements().forEach(video => {
                setPlaybackRateIfChanged(video, playbackRate);
            });
        }

        /**
         *
         * @param {number} floatVal1
         * @param {number} floatVal2
         * @param {number} [minimumDiff]
         */
        function valuesAreDifferentEnough(floatVal1, floatVal2, minimumDiff = 0.01) {
            return Math.abs(floatVal1 > floatVal2 ? floatVal1 - floatVal2 : floatVal2 - floatVal1) >= minimumDiff;
        }

        function getKnownVideoElements() {
            return knownVideoElements;
        }

        function saveVolumeLevel(volume) {
            try {
                // We save to config as a backup, in case localStorage throws an exception.
                config.mostRecentVolume = volume;
                // @todo save to chrome storage instead of localStorage.
                window.localStorage.setItem(LOCALSTORAGE_VOLUME, volume);
            } catch (ex) {
                log('failed to save volume', ex);
            }
        }

        function getSavedVolumeLevel() {
            try {
                const strVal = window.localStorage.getItem(LOCALSTORAGE_VOLUME);
                const num = typeof strVal === 'string' ? Number(strVal) : 1;
                return isNaN(num) ? config.mostRecentVolume : num;
            } catch (ex) {
                log('failed to get saved volume', ex);
            }
        }

        /**
         * They set "display: none" to hide the native controls, so we redefine their oppressive style.
         */
        function redefineWebkitMediaControlHidingCssRule() {
            const css = 'video[data-native-controls-i-set-joo-free]::-webkit-media-controls { display: flex; }';
            const id = "native-controls-i-set-joo-free";

            // Prevent from adding css multiple times per page, while re-adding it if it disappears.
            if (!document.getElementById(id)) {
                addCss(css, id);
            }
        }

        function addCss(cssCode, id) {
            const styleElement = document.createElement("style");
            id && styleElement.setAttribute("id", id);
            styleElement.appendChild(document.createTextNode(cssCode));
            document.getElementsByTagName("head")[0].appendChild(styleElement);
        }

        function removeOverlayWithInstagramPlayControl(embedRootElem) {
            // Hide the clear overlay that prevents click events from reaching the actual video element + controls.
            util.queryAll(".videoSpritePlayButton", embedRootElem).map(el => el.parentNode).forEach(hide);

            // Some more clear overlays may be hidden by this.
            getVideoCoveringButtons(embedRootElem).forEach(hide);

            // They put an img element after the video element, which totally covers the video. Hide it.
            util.queryAll("video", embedRootElem).forEach(el => {
                // Might be nice to delay this so the video can load, or remove it only after the user
                // clicks - but then the click event wont hit the video element.
                arr(el.parentNode.childNodes).filter(el => el.tagName === 'IMG').forEach(hide);
            });
        }


        /**
         * On some stories videos, they overlay the video with a "send message" textarea, but it lays on top of the video
         * player controls, making them difficult to use. So, we will identify these videos, and when found, we will reduce
         * the height of the video, so it moves upwards in its parent box, allowing the controls to be accessed. The
         * send message doesnt move - it stays fixed to the bottom of the parent.
         */
        function modifyVideoHeightIfSendMessageBoxIsBlockingVideoControls(embedRootElem) {
            // Get the textarea.
            // We need to take a few steps back up the tree before we will reach a position that can search for the textarea.
            const textarea = nthParent(embedRootElem, 4).querySelector("textarea[placeholder='Send Message']");
            if (!textarea) {
                // log('no textarea');
                return;
            }

            // This is a div that holds the "send Message" textarea and button, but it does not contain the emoticons.
            const sendMessageContainer = nthParent(textarea, 4);
            if (!sendMessageContainer) {
                log('no sendMessageContainer');
                return;
            }

            // This contains the video, the Send Message stuff, and many buttons.
            const videoAndControlsContainer = nthParent(textarea, 6);
            if (!videoAndControlsContainer) {
                log('no videoAndControlsContainer');
                return;
            }

            // Make sure they didn't change dom layout.
            const numButtons = util.queryAll('button', videoAndControlsContainer).length
            if (numButtons < 2) {
                log('numButtons < 2 for videoAndControlsContainer');
                return;
            }

            // Sometimes the video parent is taller than its parent element. The overflow is hidden, which is a problem for us because
            // the controls on the bottom of the video might not be visible. Even if they are, this overflow makes it difficult for us to
            // calculate how much to shrink the video height, so we will change the size of the parent to match. I dont know why
            // instagram does this, but maybe they don't want to change the aspect ratio too much or something.
            const video = embedRootElem.querySelector("video");
            const videoParentHeight = video.parentElement.getBoundingClientRect().height;
            const videoParentParentHeight = video.parentElement.parentElement.getBoundingClientRect().height;
            if (videoParentHeight > videoParentParentHeight) {
                log(`video parent height adjusted to ${videoParentParentHeight}px`);
                video.parentElement.style.height = videoParentParentHeight + 'px';
            }

            // We measure the height of the textarea ancestor element, because its a bit larger and contains other elements, such as the submit button.
            const distanceFromTopOfSendMessageToBottomOfParentContainer = videoAndControlsContainer.getBoundingClientRect().bottom - sendMessageContainer.getBoundingClientRect().top;
            // We move a tiny bit more to make a nice looking gap. Not sure how reliable this nudge is, but whatever.
            const videoHeightAdjustment = distanceFromTopOfSendMessageToBottomOfParentContainer + 16;

            // Now, we change the height of the video, reducing it by the height of the problematic textarea box assembly.
            // The video should move upwards, leaving a black gap below it.
            video.style.height = `calc(100% - ${videoHeightAdjustment}px)`;
            log(`video height adjusted to ${videoHeightAdjustment}px`);
        }

        function getVideoCoveringButtons(searchRoot) {
            // The button will always be very large - much bigger than a normal button since it needs to be the size of the video.
            return util.queryAll("[role=button]", searchRoot).filter(el => {
                const rect = el.getBoundingClientRect();
                return rect.width >= 200 && rect.height >= 200;
            });
        }

        /**
         * If you look in the dom, and traverse upwards from the video element, you'll soon find
         * a node which is the ancestor to both the video, and a div with role=button, and that div is big, covering the entire video.
         *
         * This func searches for that parent element. We use these as a sort of container element for the video and its
         * other controller pieces.
         *
         * @param {Element} video
         * @return {Element}
         */
        function getEstimatedVideoComponentRootElement(video) {
            for (const parent of nParents(video, 4)) {
                if (getVideoCoveringButtons(parent).length) {
                    return parent;
                }
            }
            log("couldnt find parent", video);
        }

        /**
         * Returns N number of parents, ordered with the closest elements first.
         *
         * @param {Element} elem
         * @param {number} numParents
         * @return Element[]
         */
        function nParents(elem, numParents) {
            const parents = [];
            while (elem && --numParents >= 0) {
                elem = elem.parentElement;
                elem && parents.push(elem);
            }
            return parents;
        }

        function nthParent(elem, n) {
            while (elem && --n >= 0) {
                elem = elem.parentElement;
            }
            return elem;
        }

        function remove(el) {
            el.parentNode.removeChild(el);
        }

        function attachPageVisibilityListener() {
            document.addEventListener('visibilitychange', util.recordEx('v.slfpde', startLookingForPlayerDomElement));
        }

        function startLookingForPlayerDomElement() {
            // log("startLookingForPlayerDomElement");

            // We only look for the player when the page is visible, which
            // saves battery by not running in background/idle tabs.
            if (document.hidden) {
                return;
            }

            // todo use mut obs on all pages.
            // ensureMutationObserverEnabled();

            if (isInstagramStoriesPage()) {
                // The stories GUI switches video elements very rapidly, and the videos are short duration.
                // They swap out video elements rapidly, so we need to react fast to changes, and apply our modifications
                // to each new video elements as they appear.
                // We use a MutationObserver to detect the new video elements.
                ensureMutationObserverEnabled();
            }

            util.recordEx('mv', modifyVideos)();

            // We look more rapidly on instagram because the modal dynamically creates video elements.
            setTimeout(util.recordEx('t.slfpde', startLookingForPlayerDomElement), 1000);
        }

        function ensureMutationObserverEnabled() {
            // Prevent multiple observers.
            if (mutationObserverEnabled) {
                return;
            } else {
                mutationObserverEnabled = true;
            }

            function callback(mutations) {
                log(`mutations.length = ${mutations.length}`);
                mutations.forEach(mutation => {
                    log(`mutation.addedNodes.length = ${mutation.addedNodes.length}`);
                });

                const videosAdded = [];
                mutations.forEach(mutation => {
                    mutation.addedNodes.forEach(node => {
                        const videos = util.queryAll('video', node.parentElement);
                        videosAdded.push(... videos);
                    });
                });

                log(`videosAdded.length = ${videosAdded.length}`, videosAdded);
                videosAdded.forEach(video => {
                    log('found video via mutation observer');
                    // We pass the parent node as a cheap hack to allow reuse of the modifyVideoElement func
                    modifyVideoElement(video.parentNode);
                    // todo, maybe remove some other elements.
                });
            }

            let observer = new MutationObserver(util.recordEx('mo', callback));
            observer.observe(document.body, {subtree: true, childList: true});
        }

        function isInstagramStoriesPage() {
            return window.location.pathname.toLowerCase().startsWith('/stories/');
        }

        function log() {
            if (!showLog) {
                return;
            }
            // const args = [location.href].concat(arguments);
            // const args = [location.href].concat(...arguments);
            console.log.apply(console, arguments);
        }

        this.setConfig = util.recordEx('setConfig', function setConfig(cfg) {
            // We need to be careful not to allow a 2nd call to setConfig() to cause updates to the volume of videos.
            // This could happen w/ multiple tabs open, where a user turns the vol down in 1 tab, which then gets saved to
            // chrome.storage, which then causes a storage event, which then cause the 2nd tab to receive the event and
            // call this setConfig() method, which would then update the volume of the videos. That would be annoying,
            // and prevent them from setting volume independently in 1 tab.
            // Note though, that if the page reloads, which is common, then the new volume setting from that other tab will be in effect.
            // hmmm... might be confusing behavior.
            log("new cfg", cfg);

            // Force the string to a number.
            cfg.playbackRate = Number(cfg.playbackRate);

            config = cfg;
            if (!initialized) {
                this.init();
                initialized = true;
            }

            setPlaybackRateOfPreviouslySeenVideoElements(config.playbackRate);
        });

        this.init = util.recordEx('init', function init() {
            attachPageVisibilityListener();
            startLookingForPlayerDomElement();
            redefineWebkitMediaControlHidingCssRule();
            util.recordUsageStats('initOk');
        });

    }

    const ivc = new InstagramVideoControls();
    window.instagramVideoControls = ivc;
})();