/**
 * This script injects our js and css into the webpage after it loads. It also
 * serves as a communication point between the webpage and the extension, allowing the
 * webpage to update if the user changes their settings.
 */

let showLog = false;

const MessageType = {
    SAVE_VOLUME: 1,
    USAGE_STATS: 2,
};

function init() {
    // Inject js and css.
    injectFile(chrome.extension.getURL('ic-main.js'), "script").then(() => {
        // We send a few times because sometimes the object to receive the message has constructed yet. It's rare though, so likely
        // a scheduling issues. Users w/ less cpu cores might encounter it more often.
        setTimeout(sendUpdatedConfigToWebPage, 100);
        setTimeout(sendUpdatedConfigToWebPage, 1000);
        setTimeout(sendUpdatedConfigToWebPage, 3000);
    });
}

// When the settings/config changes, we inject it into the webpage so it can update without the user needing to reload the page.
chrome.storage.onChanged.addListener(sendUpdatedConfigToWebPage);

window.addEventListener("message", function(event) {
    // Skip messages from other peoples code.
    if (!event || !event.data || !event.data.type || event.data.messageSrcAppName !== 'ic') {
        return;
    }

    const d = event.data;
    if (d && d.type === MessageType.SAVE_VOLUME && d.volumeLevel) {
        log('volume update', d);
        saveUserConfig({volumeLevel: d.volumeLevel});
    }

    if (d && d.type === MessageType.USAGE_STATS) {
        recordError(d.errorKey, d.functionName);
    }

}, false);

function log() {
    showLog && console.log.apply(console, arguments);
}

function sendUpdatedConfigToWebPage() {
    getSavedOptions().then(config => {
        // We inject a script using a data url to achieve a dynamic script.
        // Need to do this junk because of chrome content script isolated environment.
        injectFile(scriptDataUrl(scriptContentCallback(config)), "script", true);
    });
}

function removeElement(ele) {
    if (ele && ele.parentNode) {
        ele.parentNode.removeChild(ele);
    }
}

function removeElementAsync(ele) {
    setTimeout(() => {
        removeElement(ele);
    }, 15);
}

function injectFile(url, type, autoRemoveElement) {
    return new Promise(function (resolve, reject) {
        const elem = document.createElement(type);
        switch (type) {
            case 'link':
                elem.href = url;
                elem.rel  = 'stylesheet';
                elem.type = 'text/css';
                break;
            case 'script':
                elem.src = url;
                break;
            default:
                throw new Error("Unhandled switch case:" + type);
        }

        elem.setAttribute('parentExtensionId', chrome.runtime.id);

        elem.onload = function() {
            resolve();
            if (autoRemoveElement) {
                removeElementAsync(elem);
            }
        };
        elem.onerror = () => {
            reject();
            removeElementAsync(elem);
        };

        (document.head || document.documentElement).appendChild(elem);
    });
}

function scriptDataUrl(scriptContents) {
    return "data:text/javascript;base64," + btoa(scriptContents);
}

function scriptContentCallback(configArg) {
    return "window.instagramVideoControls.setConfig(" + JSON.stringify(configArg) + ")";
}

function getSavedOptions() {
    return new Promise((resolve, reject) => {
        chrome.storage.sync.get({
            rememberVolumeLevel: true,
            playbackRate: 1
        }, resolve);
    });
}

function saveUserConfig(config) {
    chrome.storage.sync.set(config, () => {

    });
}

/**
 * Increments the count for the given error key and function key.
 * You might use just 1 error key, say "stdError"
 * or maybe a few like "guiError", "saveConfigCallbackError", "backgroundScriptError"
 * etc...
 *
 * @param {string} errorKey
 * @param {string} functionName
 */
function recordError(errorKey, functionName) {
    errorKey = errorKey || 'stdError';
    functionName = functionName || 'stdFunction';
    chrome.storage.sync.get({errors: {}}, obj => {
        obj.errors = obj.errors || {};
        obj.errors[errorKey] = obj.errors[errorKey] || {};
        obj.errors[errorKey][functionName] = obj.errors[errorKey][functionName] || 0;
        obj.errors[errorKey][functionName]++;
        chrome.storage.sync.set(obj);
        log('setting sync err', obj.errors);
    });
}

init();